"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper
from dune.perftool.pymbolic.placeholder import (IndexPlaceholder,
                                                IndexPlaceholderExtraction,
                                                IndexPlaceholderRemoval,
                                                )
from dune.perftool.pdelab.geometry import GeometryMapper
from dune.perftool.generation import (domain,
                                      get_temporary_name,
                                      global_context,
                                      globalarg,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      valuearg,
                                      )

from dune.perftool.pdelab.basis import (lfs_iname,
                                        name_leaf_lfs,
                                        name_lfs,
                                        name_lfs_bound,
                                        traverse_lfs_tree,
                                        )
from dune.perftool.pdelab.quadrature import quadrature_iname
from pymbolic.primitives import Subscript, Variable


@iname
def index_sum_iname(i):
    from dune.perftool.pdelab import name_index
    return name_index(i)


_outerloop = None


def set_outerloop(v):
    global _outerloop
    _outerloop = v


def get_outerloop():
    return _outerloop


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper, GeometryMapper):
    def __init__(self, measure, subdomain_id):
        self.measure = measure
        self.subdomain_id = subdomain_id
        self.argshape = 0
        self.redinames = ()
        super(UFL2LoopyVisitor, self).__init__()

    def _assign(self, o):
        # In some corner cases we do not even need a temporary variable
        if isinstance(o, int) or isinstance(o, float):
            return o

        # Assign a name to the temporary variable we want our result in
        temp = get_temporary_name()
        temp_shape = ()

        # Determine which inames this assignment depends on and whether it should
        # be merged into the main accumulation loop. Right now we apply the following
        # procedure: All instructions that depend on all argument loop indices are
        # merged into the main loop nest. Those instructions that depend on some
        # argument loop indices but not all are merged into the kernel by fixing the
        # loop ordering of the main loop (or are pulled outside if this already happened).
        assignee = Variable(temp)
        iname_deps = self.inames
        merge_into_main_loopnest = True

        if self.rank == 2 and len(set(iname_deps)) == 1:
            if get_outerloop() is None:
                set_outerloop(iname_deps[0].number)
            if iname_deps[0].number != get_outerloop():
                merge_into_main_loopnest = False

        # Change the assignee!
        if not merge_into_main_loopnest:
            assignee_index_placeholder = IndexPlaceholderExtraction()(o).pop()
            assignee_index = IndexPlaceholderRemoval(duplicate_inames=True)(assignee_index_placeholder)
            assignee = Subscript(assignee, (assignee_index,))
            temp_shape = (name_lfs_bound(name_leaf_lfs(assignee_index_placeholder.element, assignee_index_placeholder.restriction)),)

        # Now introduce duplicate inames for the argument loop if necessary
        replaced_iname_deps = [IndexPlaceholderRemoval(duplicate_inames=not merge_into_main_loopnest, wrap_in_variable=False)(i) for i in iname_deps]
        replaced_expr = IndexPlaceholderRemoval(duplicate_inames=not merge_into_main_loopnest)(o)

        # Now we assign this expression to a new temporary variable
        insn_id = instruction(assignee=assignee,
                              expression=replaced_expr,
                              forced_iname_deps=frozenset({i for i in replaced_iname_deps}).union(frozenset({quadrature_iname()})),
                              forced_iname_deps_is_final=True,
                              )

        # Actually, if we have a cache hit, we need to change our temporary
        from dune.perftool.generation import retrieve_cache_items
        temp = filter(lambda i: i.id == insn_id, retrieve_cache_items("instruction"))[0].assignee_name

        retvar = Variable(temp)
        if not merge_into_main_loopnest:
            retvar_index = IndexPlaceholderRemoval()(assignee_index_placeholder)
            retvar = Subscript(retvar, (retvar_index,))

        # Now that we know its exact name, declare the temporary
        temporary_variable(temp, shape=temp_shape)

        return retvar

    def __call__(self, o):
        # Initialize the local function spaces that we might need for this term
        # We therefore gather a list of modified trial functions too.
        from dune.perftool.ufl.modified_terminals import extract_modified_arguments
        test_ma = extract_modified_arguments(o)
        trial_ma = extract_modified_arguments(o, trialfunction=True, testfunction=False)

        # All test and trial functions on boundary integrals are technically restricted
        import itertools
        for ma in itertools.chain(test_ma, trial_ma):
            if self.measure == 'exterior_facet':
                ma.restriction = Restriction.NEGATIVE

            traverse_lfs_tree(ma)

        # Determine the rank of the term
        self.rank = len(test_ma)

        # And initialize a list of found inames
        self.inames = []

        # First we do the tree traversal to get a pymbolic expression representing this expression
        pymbolic_expr = self.call(o)

        # It may happen that an entire accumulation term vanishes. We do nothing in that case
        if pymbolic_expr == 0:
            return

        # We assign the result to a temporary variable to ease things a bit
        if not isinstance(pymbolic_expr, Variable):
            pymbolic_expr = self._assign(pymbolic_expr)

        # Transform the IndexPlaceholders into real inames
        self.inames = [IndexPlaceholderRemoval(wrap_in_variable=False)(i) for i in self.inames]

        # Collect the arguments for the accumulate function
        accumargs = [None] * (2 * len(test_ma))
        residual_shape = [None] * len(test_ma)
        arg_restr = [None] * len(test_ma)

        for ma in test_ma:
            count = ma.argexpr.number()
            if len(test_ma) == 2:
                icount = (count + 1) % 2
            else:
                icount = 0

            # Extract the subelement of the (mixed) finite element
            from ufl.functionview import select_subelement
            subel = select_subelement(ma.argexpr.ufl_element(), ma.component)

            # And generate a local function space for it!
            lfs = name_lfs(ma.argexpr.ufl_element(), ma.restriction, ma.component)
            if len(subel.value_shape()) != 0:
                from dune.perftool.pdelab.geometry import dimension_iname
                from dune.perftool.pdelab.basis import lfs_child
                lfs = lfs_child(lfs, dimension_iname(count=count, context='arg'))
                residual_shape[icount] = name_lfs_bound(name_leaf_lfs(subel.sub_elements()[0], ma.restriction))
            else:
                residual_shape[icount] = name_lfs_bound(lfs)

            lfsi = lfs_iname(subel, ma.restriction, count=count)

            accumargs[2 * icount] = lfs
            accumargs[2 * icount + 1] = lfsi

            arg_restr[icount] = ma.restriction

        from dune.perftool.pdelab.argument import name_accumulation_variable
        accumvar = name_accumulation_variable(arg_restr)

        from dune.perftool.pdelab.quadrature import name_factor
        factor = name_factor()

        # Generate the code snippet for this accumulation instruction
        code = "{}.accumulate({}, {}*{});".format(accumvar,
                                                  ", ".join(accumargs),
                                                  pymbolic_expr.name,
                                                  factor,
                                                  )

        # Maybe wrap this instruction into a condiditional. This mostly happens with mixed boundary conditions
        if self.subdomain_id not in ['everywhere', 'otherwise']:
            # We need to reconstruct the subdomain_data parameter of the measure
            # I am *totally* confused as to why this information is not at hand anyway,
            # but conversation with Martin pointed me to dolfin.fem.assembly where this
            # is done in preprocessing with the limitation of only one possible type of
            # modified measure per integral type.

            # Get the original form and inspect the present measures
            from dune.perftool.generation import get_global_context_value
            original_form = get_global_context_value("formdata").original_form

            sd = original_form.subdomain_data()
            assert len(sd) == 1
            subdomains, = list(sd.values())
            domain, = list(sd.keys())
            for k in list(subdomains.keys()):
                if subdomains[k] is None:
                    del subdomains[k]

            # Finally extract the original subdomain_data (which needs to be present!)
            assert self.measure in subdomains
            subdomain_data = subdomains[self.measure]

            # Determine the name of the parameter function
            name = get_global_context_value("namedata")[id(subdomain_data)]

            # Trigger the generation of code for this thing in the parameter class
            from dune.perftool.pdelab.parameter import intersection_parameter_function
            intersection_parameter_function(name, subdomain_data, t='int')

            code = "if ({} == {})\n  {}".format(name, self.subdomain_id, code)

        # Finally, issue the instruction
        instruction(code=code,
                    read_variables=frozenset({factor, pymbolic_expr.name}),
                    forced_iname_deps=frozenset(self.inames).union(frozenset({quadrature_iname()})),
                    forced_iname_deps_is_final=True,
                    )

    def argument(self, o):
        # Correct the restriction on boundary integrals
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        # Select the correct subtree of the finite element
        from ufl.functionview import select_subelement
        element = select_subelement(o.ufl_element(), self.component)
        leaf_element = element

        # Have the issued instruction depend on the iname for this localfunction space
        self.inames.append(IndexPlaceholder(element, restriction, o.number()))

        # Now treat the case of this being a vector finite element
        if element.num_sub_elements() > 0:
            # I cannot handle general mixed elements here...
            from ufl import VectorElement
            assert isinstance(element, VectorElement)

            # Determine whether this is a non-scalar subargument. This information is later
            # used by the Indexed-Mapper to make some indices loop indices!
            self.argshape = len(element.value_shape())

            # If this is a vector element, we need add an additional accumulation loop iname
            from dune.perftool.pdelab.geometry import dimension_iname
            for i in range(self.argshape):
                self.inames.append(dimension_iname(count=i, context='arg'))

            # For the purpose of basis evaluation, we need to take the leaf element
            leaf_element = element.sub_elements()[0]

        if self.grad:
            from dune.perftool.pdelab.argument import name_testfunction_gradient
            return Subscript(Variable(name_testfunction_gradient(leaf_element, restriction)), (IndexPlaceholder(element, restriction, o.number()),))
        else:
            from dune.perftool.pdelab.argument import name_testfunction
            return Subscript(Variable(name_testfunction(leaf_element, restriction)), (IndexPlaceholder(element, restriction, o.number()),))

    def coefficient(self, o):
        # If this is a trialfunction, we do something entirely different
        if o.count() == 0:
            # Correct the restriction on boundary integrals
            restriction = self.restriction
            if self.measure == 'exterior_facet':
                restriction = Restriction.NEGATIVE

            if self.grad:
                from dune.perftool.pdelab.argument import name_trialfunction_gradient
                return Variable(name_trialfunction_gradient(o.ufl_element(), restriction, self.component))
            else:
                from dune.perftool.pdelab.argument import name_trialfunction
                return Variable(name_trialfunction(o.ufl_element(), restriction, self.component))

        # so this is a parameter function
        else:
            # We expect all coefficients to be of type Expression!
            from dune.perftool.ufl.execution import Expression
            assert isinstance(o, Expression)

            # Determine the name of the parameter function
            from dune.perftool.generation import get_global_context_value
            name = get_global_context_value("namedata")[id(o)]

            # Trigger the generation of code for this thing in the parameter class
            from dune.perftool.pdelab.parameter import (cell_parameter_function,
                                                        intersection_parameter_function,
                                                        )
            if o.on_intersection:
                intersection_parameter_function(name, o)
            else:
                cell_parameter_function(name, o, self.restriction)

            # And return a symbol
            return Variable(name)

    def indexed(self, o):
        """ Although indexed is something that we want to handle in the pymbolic mapper we
        need to handle it here, as we have to give special treatment to indices into vector
        elements: they become a loop index! """
        # Set the generated index for the argument handler to have it available
        self.last_index = self.call(o.ufl_operands[1])
        aggr = self.call(o.ufl_operands[0])

        use_indices = self.last_index[self.argshape:]
        self.argshape = 0
        if isinstance(aggr, Subscript):
            return Subscript(aggr.aggregate, aggr.index + use_indices)
        else:
            return Subscript(aggr, use_indices)

    def index_sum(self, o):
        # Get the iname for the reduction index
        ind = o.ufl_operands[1][0]
        self.redinames = self.redinames + (index_sum_iname(ind),)
        shape = o.ufl_operands[0].ufl_index_dimensions[0]
        from dune.perftool.pdelab import name_index
        domain(name_index(ind), shape)

        # If the left operand is an index sum to, we do it in one reduction
        from ufl.classes import IndexSum
        if isinstance(o.ufl_operands[0], IndexSum):
            return self.call(o.ufl_operands[0])
        else:
            from loopy import Reduction

            oldinames = self.inames
            self.inames = []

            # Recurse to get the summation expression
            term = self.call(o.ufl_operands[0])
            ret = self._assign(Reduction("sum", self.redinames, term))

            self.inames = self.inames + oldinames

            # Reset the reduction inames for future indexsums
            self.redinames = ()

            return ret


def transform_accumulation_term(term, measure, subdomain_id):
    return UFL2LoopyVisitor(measure, subdomain_id)(term)
